from __future__ import annotations

import importlib.util
import os
import re
import tempfile
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path

import numpy as np
import pytest
from numpy.typing import DTypeLike

import snaphu

# Maximum value representable by 32-bit unsigned integer type.
UINT32_MAX = np.iinfo(np.uint32).max


def has_rasterio() -> bool:
    """Check if `rasterio` can be imported."""
    return importlib.util.find_spec("rasterio") is not None


@contextmanager
def make_geotiff_raster(
    fp: str | os.PathLike[str],
    shape: tuple[int, int] = (1024, 512),
    dtype: DTypeLike = np.int32,
    epsg: int = 4326,
    extents: tuple[float, float, float, float] = (0.0, 0.0, 1.0, 1.0),
    nodata: float | None = None,
) -> Generator[snaphu.io.Raster, None, None]:
    """
    Make a dummy GeoTiff raster for testing.

    Upon exiting the context manager, the raster dataset is closed.

    Parameters
    ----------
    fp : path-like
        File system path of the dataset. The parent directory must exist.
    shape : (int, int), optional
        The raster shape (height and width).
    dtype : data-type, optional
        The raster data-type. Must be convertible to `numpy.dtype`.
    epsg : int, optional
        An EPSG code defining the coordinate reference system (CRS) of the raster.
    extents : (float, float, float, float), optional
        The geospatial extents of the raster image, in coordinates defined by the CRS
        represented by the `epsg` code, in the following order: West, South, East,
        North.
    nodata : float or None, optional
        Defines the pixel value to be interpreted as not valid data.

    Yields
    ------
    raster : snaphu.io.Raster
        The raster.
    """
    import rasterio

    height, width = shape
    crs = rasterio.crs.CRS.from_epsg(epsg)
    transform = rasterio.transform.from_bounds(*extents, width=width, height=height)
    with snaphu.io.Raster.create(
        fp,
        width=width,
        height=height,
        dtype=dtype,
        crs=crs,
        transform=transform,
        nodata=nodata,
        driver="GTiff",
    ) as raster:
        yield raster


@contextmanager
def make_temp_geotiff_raster() -> Generator[snaphu.io.Raster, None, None]:
    """
    Make a temporary dummy GeoTiff raster dataset for testing.

    Upon exiting the context manager, the raster dataset is closed and the file is
    removed from the file system.

    Yields
    ------
    raster : snaphu.io.Raster
        The raster.
    """
    with (
        tempfile.NamedTemporaryFile(suffix=".tif") as file_,
        make_geotiff_raster(file_.name) as raster,
    ):
        yield raster


@pytest.mark.skipif(not has_rasterio(), reason="requires rasterio package")
class TestRaster:
    @pytest.fixture(scope="class")
    def geotiff_raster(self) -> Generator[snaphu.io.Raster, None, None]:
        with make_temp_geotiff_raster() as raster:
            yield raster

    def test_is_input_dataset(self, geotiff_raster: snaphu.io.Raster):
        assert isinstance(geotiff_raster, snaphu.io.InputDataset)

    def test_is_output_dataset(self, geotiff_raster: snaphu.io.Raster):
        assert isinstance(geotiff_raster, snaphu.io.OutputDataset)

    def test_open_raster(self):
        with tempfile.NamedTemporaryFile(suffix=".tif") as file_:
            with make_geotiff_raster(file_.name):
                pass

            assert Path(file_.name).is_file()

            with snaphu.io.Raster(file_.name) as raster:
                assert raster.mode == "r"

            with snaphu.io.Raster(file_.name, mode="r+") as raster:
                assert raster.mode == "r+"

    def test_create_like(self, geotiff_raster: snaphu.io.Raster):
        with (
            tempfile.NamedTemporaryFile(suffix=".tif") as file_,
            snaphu.io.Raster.create(file_.name, like=geotiff_raster) as new_raster,
        ):
            assert new_raster.shape == geotiff_raster.shape
            assert new_raster.dtype == geotiff_raster.dtype
            assert new_raster.driver == geotiff_raster.driver
            assert new_raster.crs == geotiff_raster.crs
            assert new_raster.transform == geotiff_raster.transform

            assert new_raster.band == 1
            assert new_raster.dataset.count == 1

    def test_create_like_override(self, geotiff_raster: snaphu.io.Raster):
        with tempfile.TemporaryDirectory() as dir_:
            fp = Path(dir_) / "raster"
            with snaphu.io.Raster.create(
                fp,
                like=geotiff_raster,
                dtype="f8",
                driver="ENVI",
            ) as new_raster:
                assert new_raster.shape == geotiff_raster.shape
                assert new_raster.crs == geotiff_raster.crs
                assert new_raster.transform == geotiff_raster.transform
                assert new_raster.dtype == np.float64
                assert new_raster.driver == "ENVI"

    def test_shape(self, geotiff_raster: snaphu.io.Raster):
        shape = (1024, 512)
        assert geotiff_raster.shape == shape
        assert geotiff_raster.height == shape[0]
        assert geotiff_raster.width == shape[1]

    def test_dtype(self, geotiff_raster: snaphu.io.Raster):
        assert isinstance(geotiff_raster.dtype, np.dtype)
        assert geotiff_raster.dtype == np.int32

    def test_ndim(self, geotiff_raster: snaphu.io.Raster):
        assert geotiff_raster.ndim == 2

    def test_band(self, geotiff_raster: snaphu.io.Raster):
        assert geotiff_raster.band == 1

    def test_mode(self, geotiff_raster: snaphu.io.Raster):
        assert geotiff_raster.mode == "w+"

    def test_driver(self, geotiff_raster: snaphu.io.Raster):
        assert geotiff_raster.driver == "GTiff"

    def test_crs(self, geotiff_raster: snaphu.io.Raster):
        import rasterio

        assert geotiff_raster.crs == rasterio.crs.CRS.from_epsg(4326)

    def test_transform(self, geotiff_raster: snaphu.io.Raster):
        import rasterio

        height, width = 1024, 512
        transform = rasterio.transform.from_bounds(0.0, 0.0, 1.0, 1.0, width, height)
        assert geotiff_raster.transform == transform

    @pytest.mark.parametrize(
        ("dtype", "nodata"),
        [
            (np.int32, -999.0),
            (np.uint32, UINT32_MAX),
            (np.float64, 123.456),
            (np.complex64, 0.0),
            (np.float32, None),
        ],
    )
    def test_nodata(self, dtype: DTypeLike, nodata: float | None):
        with (
            tempfile.NamedTemporaryFile(suffix=".tif") as file_,
            make_geotiff_raster(file_.name, dtype=dtype, nodata=nodata) as raster,
        ):
            assert raster.nodata == nodata

    def test_open_closed(self):
        with make_temp_geotiff_raster() as raster:
            assert not raster.closed
        assert raster.closed

    def test_arraylike(self, geotiff_raster: snaphu.io.Raster):
        arr = np.asarray(geotiff_raster)
        assert arr.shape == geotiff_raster.shape
        assert arr.dtype == geotiff_raster.dtype

    def test_setitem_getitem_roundtrip(self, geotiff_raster: snaphu.io.Raster):
        data = np.arange(20, dtype=np.int32).reshape(4, 5)
        idx = np.s_[100:104, 200:205]
        geotiff_raster[idx] = data
        out = geotiff_raster[idx]
        np.testing.assert_array_equal(out, data)

    def test_repr(self, geotiff_raster: snaphu.io.Raster):
        dataset = repr(geotiff_raster.dataset)
        band = repr(geotiff_raster.band)
        assert repr(geotiff_raster) == f"Raster(dataset={dataset}, band={band})"

    def test_bad_mode(self):
        with tempfile.NamedTemporaryFile(suffix=".tif") as file_:
            fp = file_.name
            with make_geotiff_raster(fp):
                pass

            regex = re.compile(r"unsupported mode 'q', must be one of \{'r', 'r\+'\}")
            with pytest.raises(ValueError, match=regex):
                snaphu.io.Raster(fp, mode="q")

    def test_bad_band(self):
        with tempfile.NamedTemporaryFile(suffix=".tif") as file_:
            fp = file_.name
            with make_geotiff_raster(fp):
                pass

            regex = re.compile(
                r"band index 2 out of range: dataset contains 1 raster band\(s\)"
            )
            with pytest.raises(IndexError, match=regex):
                snaphu.io.Raster(fp, band=2)



import re

import snaphu


def test_version_pep440():
    # Check that the version string is PEP440-compliant.
    # https://peps.python.org/pep-0440/#appendix-b-parsing-version-strings-with-regular-expressions
    regex = re.compile(
        r"""
        v?
        (?:
            (?:(?P<epoch>[0-9]+)!)?                           # epoch
            (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
            (?P<pre>                                          # pre-release
                [-_\.]?
                (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
                [-_\.]?
                (?P<pre_n>[0-9]+)?
            )?
            (?P<post>                                         # post release
                (?:-(?P<post_n1>[0-9]+))
                |
                (?:
                    [-_\.]?
                    (?P<post_l>post|rev|r)
                    [-_\.]?
                    (?P<post_n2>[0-9]+)?
                )
            )?
            (?P<dev>                                          # dev release
                [-_\.]?
                (?P<dev_l>dev)
                [-_\.]?
                (?P<dev_n>[0-9]+)?
            )?
        )
        (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
        """,
        re.VERBOSE | re.IGNORECASE,
    )
    assert regex.fullmatch(snaphu.__version__)


def test_version_tuple():
    # `__version_tuple__` should contain exactly 3 nonnegative integer release numbers,
    # plus optional string(s) representing pre/post/dev release segments and/or local
    # version label.
    major, minor, patch, *extra = snaphu.__version_tuple__
    assert all(isinstance(n, int) for n in [major, minor, patch])
    assert (major, minor, patch) >= (0, 0, 0)
    assert all(isinstance(s, str) for s in extra)



# The `_version` module is auto-generated by setuptools_scm at install time.
from . import io
from ._snaphu import get_snaphu_version
from ._unwrap import unwrap
from ._version import __version__, __version_tuple__

__all__ = [
    "get_snaphu_version",
    "io",
    "unwrap",
    "__version__",
    "__version_tuple__",
]



import numpy as np
import pytest

import snaphu


class TestUnwrap:
    @pytest.mark.parametrize("cost", ["defo", "smooth"])
    @pytest.mark.parametrize("init", ["mst", "mcf"])
    def test_unwrapped_phase(self, cost: str, init: str):
        # Simulate interferogram containing a diagonal phase ramp with multiple fringes.
        y, x = np.ogrid[-3:3:512j, -3:3:512j]
        phase = np.pi * (x + y)
        igram = np.exp(1j * phase)

        # Sample coherence for an interferogram with no noise.
        corr = np.ones(igram.shape, dtype=np.float32)

        # Unwrap.
        unw, _ = snaphu.unwrap(igram, corr, nlooks=1.0, cost=cost, init=init)

        # The unwrapped phase may differ from the true phase by a fixed integer multiple
        # of 2pi.
        mean_diff = np.mean(unw - phase)
        offset = 2.0 * np.pi * np.round(mean_diff / (2.0 * np.pi))
        np.testing.assert_allclose(unw, phase + offset, atol=1e-3)

    def test_mask(self):
        # Simulate interferogram containing a diagonal phase ramp with multiple fringes.
        y, x = np.ogrid[-3:3:512j, -3:3:512j]
        phase = np.pi * (x + y)
        igram = np.exp(1j * phase)

        # Sample coherence for an interferogram with no noise.
        corr = np.ones(igram.shape, dtype=np.float32)

        # Create a binary mask of valid samples.
        mask = np.zeros(igram.shape, dtype=np.bool_)
        mask[128:-128] = True

        # Unwrap.
        unw, conncomp = snaphu.unwrap(igram, corr, nlooks=1.0, mask=mask)

        # The unwrapped phase may differ from the true phase by a fixed integer multiple
        # of 2pi.
        mean_diff = np.mean(unw[mask] - phase[mask])
        offset = 2.0 * np.pi * np.round(mean_diff / (2.0 * np.pi))
        np.testing.assert_allclose(unw[mask], phase[mask] + offset, atol=1e-3)

        # There should be a single connected component (labeled 1) that contains all of
        # the valid pixels and none of the invalid pixels.
        np.testing.assert_array_equal(conncomp, mask.astype(np.int32))

    @pytest.mark.parametrize("nan", [np.nan, 1j * np.nan])
    def test_nans(self, nan: complex):
        # Simulate interferogram containing a diagonal phase ramp with multiple fringes.
        y, x = np.ogrid[-3:3:512j, -3:3:512j]
        phase = np.pi * (x + y)
        igram = np.exp(1j * phase)

        # Sample coherence for an interferogram with no noise.
        corr = np.ones(igram.shape, dtype=np.float32)

        # Create a binary mask of valid samples.
        mask = np.zeros(igram.shape, dtype=np.bool_)
        mask[128:-128] = True

        # Set masked-out interferogram pixels to NaN.
        igram[~mask] = nan

        # Unwrap.
        unw, conncomp = snaphu.unwrap(igram, corr, nlooks=1.0)

        # The unwrapped phase may differ from the true phase by a fixed integer multiple
        # of 2pi.
        mean_diff = np.mean(unw[mask] - phase[mask])
        offset = 2.0 * np.pi * np.round(mean_diff / (2.0 * np.pi))
        np.testing.assert_allclose(unw[mask], phase[mask] + offset, atol=1e-3)

        # There should be a single connected component (labeled 1) that contains all of
        # the valid pixels and none of the invalid pixels.
        np.testing.assert_array_equal(conncomp, mask.astype(np.int32))

    @pytest.mark.parametrize("nproc", [1, 2, -1])
    def test_tiling(self, nproc: int):
        # Simulate interferogram containing a diagonal phase ramp with multiple fringes.
        y, x = np.ogrid[-6:6:1024j, -6:6:1024j]
        phase = np.pi * (x + y)
        igram = np.exp(1j * phase)

        # Sample coherence for an interferogram with no noise.
        corr = np.ones(igram.shape, dtype=np.float32)

        # Unwrap.
        unw, _ = snaphu.unwrap(
            igram,
            corr,
            nlooks=1.0,
            ntiles=(2, 2),
            tile_overlap=(128, 128),
            nproc=nproc,
        )

        # The unwrapped phase may differ from the true phase by a fixed integer multiple
        # of 2pi.
        mean_diff = np.mean(unw - phase)
        offset = 2.0 * np.pi * np.round(mean_diff / (2.0 * np.pi))
        np.testing.assert_allclose(unw, phase + offset, atol=1e-2)

    def test_regrow_conncomps(self):
        # Simulate interferogram containing a diagonal phase ramp with multiple fringes.
        y, x = np.ogrid[-3:3:512j, -3:3:512j]
        phase = np.pi * (x + y)
        igram = np.exp(1j * phase)

        # Create a binary mask containing a single U-shaped region of valid data that
        # spans the 4 quadrants of the interferogram.
        mask = np.zeros(igram.shape, dtype=np.bool_)
        mask = np.zeros(igram.shape, dtype=np.bool_)
        mask[:, :128] = True
        mask[-128:, :] = True
        mask[:, -128:] = True

        # Sample coherence for an interferogram with no noise.
        corr = np.ones(igram.shape, dtype=np.float32)

        # Unwrap using a 2x2 grid of tiles. Without regrowing connected components, this
        # would result in 4 distinct connected component labels (one per tile).
        _, conncomp = snaphu.unwrap(
            igram,
            corr,
            nlooks=1.0,
            mask=mask,
            ntiles=(2, 2),
            tile_overlap=(64, 64),
            regrow_conncomps=True,
        )

        # There should be a single connected component (labeled 1) that contains all of
        # the valid pixels and none of the invalid pixels.
        np.testing.assert_array_equal(conncomp, mask.astype(np.int32))

    def test_shape_mismatch(self):
        igram = np.empty(shape=(128, 128), dtype=np.complex64)
        corr = np.empty(shape=(128, 129), dtype=np.float32)
        pattern = (
            "^shape mismatch: corr and igram must have the same shape, instead got"
            r" corr.shape=\(128, 129\) and igram.shape=\(128, 128\)$"
        )
        with pytest.raises(ValueError, match=pattern):
            snaphu.unwrap(igram, corr, nlooks=100.0)

    def test_bad_igram_dtype(self):
        shape = (128, 128)
        igram = np.empty(shape, dtype=np.float64)
        corr = np.empty(shape, dtype=np.float32)
        pattern = r"^igram must be a complex-valued array, instead got dtype=float64$"
        with pytest.raises(TypeError, match=pattern):
            snaphu.unwrap(igram, corr, nlooks=100.0)

    def test_bad_corr_dtype(self):
        shape = (128, 128)
        igram = np.empty(shape, dtype=np.complex64)
        corr = np.empty(shape, dtype=np.complex64)
        pattern = r"^corr must be a real-valued array, instead got dtype=complex64$"
        with pytest.raises(TypeError, match=pattern):
            snaphu.unwrap(igram, corr, nlooks=100.0)

    def test_bad_nlooks(self):
        shape = (128, 128)
        igram = np.empty(shape, dtype=np.complex64)
        corr = np.empty(shape, dtype=np.float32)
        pattern = "^nlooks must be >= 1, instead got 0$"
        with pytest.raises(ValueError, match=pattern):
            snaphu.unwrap(igram, corr, nlooks=0)

    def test_bad_cost(self):
        shape = (128, 128)
        igram = np.empty(shape, dtype=np.complex64)
        corr = np.empty(shape, dtype=np.float32)
        pattern = r"^cost mode must be in \{.*\}, instead got 'asdf'$"
        with pytest.raises(ValueError, match=pattern):
            snaphu.unwrap(igram, corr, nlooks=100.0, cost="asdf")

    def test_bad_init(self):
        shape = (128, 128)
        igram = np.empty(shape, dtype=np.complex64)
        corr = np.empty(shape, dtype=np.float32)
        pattern = r"^init method must be in \{.*\}, instead got 'asdf'$"
        with pytest.raises(ValueError, match=pattern):
            snaphu.unwrap(igram, corr, nlooks=100.0, init="asdf")

    def test_bad_ntiles(self):
        shape = (128, 128)
        igram = np.empty(shape, dtype=np.complex64)
        corr = np.empty(shape, dtype=np.float32)

        pattern = r"^ntiles must be a pair of ints, instead got ntiles=\(1, 2, 3\)$"
        with pytest.raises(ValueError, match=pattern):
            snaphu.unwrap(igram, corr, nlooks=100.0, ntiles=(1, 2, 3))  # type: ignore[arg-type]

        pattern = (
            r"^ntiles may not contain negative or zero values, got ntiles=\(1, 0\)$"
        )
        with pytest.raises(ValueError, match=pattern):
            snaphu.unwrap(igram, corr, nlooks=100.0, ntiles=(1, 0))

    def test_bad_tile_overlap(self):
        shape = (128, 128)
        igram = np.empty(shape, dtype=np.complex64)
        corr = np.empty(shape, dtype=np.float32)

        pattern = (
            r"^tile_overlap must be an int or pair of ints, instead got"
            r" tile_overlap=\(1, 2, 3\)$"
        )
        with pytest.raises(ValueError, match=pattern):
            snaphu.unwrap(igram, corr, nlooks=100.0, tile_overlap=(1, 2, 3))  # type: ignore[arg-type]

        pattern = (
            r"^tile_overlap may not contain negative values, got"
            r" tile_overlap=\(0, -1\)$"
        )
        with pytest.raises(ValueError, match=pattern):
            snaphu.unwrap(igram, corr, nlooks=100.0, tile_overlap=(0, -1))



from __future__ import annotations

import os
from collections.abc import Mapping
from contextlib import AbstractContextManager
from typing import Any, TypeVar

import numpy as np
import rasterio
from numpy.typing import DTypeLike

from ._dataset import InputDataset, OutputDataset

__all__ = [
    "Raster",
]


RasterT = TypeVar("RasterT", bound="Raster")


class Raster(InputDataset, OutputDataset, AbstractContextManager["Raster"]):
    """
    A single raster band in a GDAL-compatible dataset containing one or more bands.

    `Raster` provides a convenient interface for using SNAPHU to unwrap ground-projected
    interferograms in raster formats supported by the Geospatial Data Abstraction
    Library (GDAL). It acts as a thin wrapper around a Rasterio dataset and a band
    index, providing NumPy-like access to the underlying raster data.

    Data access is performed lazily -- the raster contents are not stored in memory
    unless/until they are explicitly accessed by an indexing operation.

    `Raster` objects must be closed after use in order to ensure that any data written
    to them is flushed to disk and any associated file objects are closed. The `Raster`
    class implements Python's context manager protocol, which can be used to reliably
    ensure that the raster is closed upon exiting the context manager.
    """

    def __init__(
        self,
        fp: str | os.PathLike[str],
        mode: str = "r",
        band: int = 1,
        driver: str | None = None,
    ):
        """
        Open an existing raster dataset.

        Parameters
        ----------
        fp : str or path-like
            File system path or URL of the local or remote dataset.
        mode : {'r', 'r+'}, optional
            The file access mode -- 'r' for read-only (the default) or 'r+' for
            read/write access.
        band : int, optional
            The (1-based) band index of the raster band. Defaults to 1.
        driver : str or None, optional
            Raster format driver name. If None, registered drivers will be tried
            sequentially until a match is found. Defaults to None.
        """
        # Check `mode` argument.
        if mode in {"w", "w+"}:
            clsname = type(self).__name__
            errmsg = (
                f"{clsname}.__init__() only supports 'r' and 'r+' modes, use"
                f" {clsname}.create() for dataset creation"
            )
            raise ValueError(errmsg)
        if mode not in {"r", "r+"}:
            errmsg = f"unsupported mode {mode!r}, must be one of {{'r', 'r+'}}"
            raise ValueError(errmsg)

        # Open the dataset.
        dataset = rasterio.open(fp, mode=mode, driver=driver)

        # Check that `band` is a valid band index in the dataset.
        nbands = dataset.count
        if not (1 <= band <= nbands):
            errmsg = (
                f"band index {band} out of range: dataset contains {nbands} raster"
                " band(s)"
            )
            raise IndexError(errmsg)

        self._dataset = dataset
        self._band = band

    @classmethod
    def create(
        cls: type[RasterT],
        fp: str | os.PathLike[str],
        width: int | None = None,
        height: int | None = None,
        dtype: DTypeLike | None = None,
        driver: str | None = None,
        crs: str | Mapping[str, str] | rasterio.crs.CRS | None = None,
        transform: rasterio.transform.Affine | None = None,
        nodata: float | None = None,
        *,
        like: Raster | None = None,
        **kwargs: Any,
    ) -> RasterT:
        """
        Create a new single-band raster dataset.

        If another raster is passed via the `like` argument, the new dataset will
        inherit the shape, data-type, driver, coordinate reference system (CRS), and
        geotransform of the reference raster. Driver-specific dataset creation options
        such as chunk size and compression flags may also be inherited.

        All other arguments take precedence over `like` and may be used to override
        attributes of the reference raster when creating the new raster.

        Parameters
        ----------
        fp : str or path-like
            File system path or URL of the local or remote dataset.
        width, height : int or None, optional
            The numbers of columns and rows of the raster dataset. Required if `like` is
            not specified. Otherwise, if None, the new dataset is created with the same
            width/height as `like`. Defaults to None.
        dtype : data-type or None, optional
            Data-type of the raster dataset's elements. Must be convertible to a
            `numpy.dtype` object and must correspond to a valid GDAL datatype. Required
            if `like` is not specified. Otherwise, if None, the new dataset is created
            with the same data-type as `like`. Defaults to None.
        driver : str or None, optional
            Raster format driver name. If None, the method will attempt to infer the
            driver from the file extension. Defaults to None.
        crs : str, dict, rasterio.crs.CRS, or None; optional
            The coordinate reference system. If None, the CRS of `like` will be used, if
            available, otherwise the raster will not be georeferenced. Defaults to None.
        transform : rasterio.transform.Affine or None, optional
            Affine transformation mapping the pixel space to geographic space. If None,
            the geotransform of `like` will be used, if available, otherwise the default
            transform will be used. Defaults to None.
        nodata : float or None, optional
            Defines the pixel value to be interpreted as not valid data. If None, the
            nodata value of `like` will be used, if available, otherwise no nodata value
            will be populated. Only real-valued nodata values are supported, even if the
            raster data is complex-valued. Defaults to None.
        like : Raster or None, optional
            An optional reference raster. If not None, the new raster will be created
            with the same metadata (shape, data-type, driver, CRS/geotransform, etc) as
            the reference raster. All other arguments will override the corresponding
            attribute of the reference raster. Defaults to None.
        **kwargs : dict, optional
            Additional driver-specific creation options passed to `rasterio.open`.
        """
        if like is not None:
            kwargs = like.dataset.profile | kwargs

        if width is not None:
            kwargs["width"] = width
        if height is not None:
            kwargs["height"] = height
        if dtype is not None:
            kwargs["dtype"] = np.dtype(dtype)
        if driver is not None:
            kwargs["driver"] = driver
        if crs is not None:
            kwargs["crs"] = crs
        if transform is not None:
            kwargs["transform"] = transform
        if nodata is not None:
            kwargs["nodata"] = nodata

        # Always create a single-band dataset, even if `like` was part of a multi-band
        # dataset.
        kwargs["count"] = 1

        # Create the new single-band dataset.
        dataset = rasterio.open(fp, mode="w+", **kwargs)

        # XXX We need this gross hack in order to bypass calling `__init__` (which only
        # supports opening existing datasets).
        raster = cls.__new__(cls)
        raster._dataset = dataset  # noqa: SLF001
        raster._band = 1  # noqa: SLF001

        return raster

    @property
    def dtype(self) -> np.dtype:
        return np.dtype(self.dataset.dtypes[self.band - 1])

    @property
    def height(self) -> int:
        """int : The number of rows in the raster."""  # noqa: D403
        return self.dataset.height  # type: ignore[no-any-return]

    @property
    def width(self) -> int:
        """int : The number of columns in the raster."""  # noqa: D403
        return self.dataset.width  # type: ignore[no-any-return]

    @property
    def shape(self) -> tuple[int, int]:
        return self.height, self.width

    @property
    def ndim(self) -> int:
        return 2

    @property
    def dataset(self) -> rasterio.io.DatasetReader | rasterio.io.DatasetWriter:
        """
        rasterio.io.DatasetReader or rasterio.io.DatasetWriter : The underlying dataset.
        """  # noqa: D200
        return self._dataset

    @property
    def band(self) -> int:
        """int : Band index (1-based)."""  # noqa: D403
        return self._band

    @property
    def mode(self) -> str:
        """str : File access mode."""  # noqa: D403
        return self.dataset.mode  # type: ignore[no-any-return]

    @property
    def driver(self) -> str:
        """str : Raster format driver name."""  # noqa: D403
        return self.dataset.driver  # type: ignore[no-any-return]

    @property
    def crs(self) -> rasterio.crs.CRS:
        """rasterio.crs.CRS : The dataset's coordinate reference system."""
        return self.dataset.crs

    @property
    def transform(self) -> rasterio.transform.Affine:
        """
        rasterio.transform.Affine : The dataset's georeferencing transformation matrix.

        This transform maps pixel row/column coordinates to coordinates in the dataset's
        coordinate reference system.
        """
        return self.dataset.transform

    @property
    def nodata(self) -> float | None:
        """
        float or None : The raster's nodata value (may be unset).

        The raster's nodata value, or None if no nodata value was set.
        """
        return self.dataset.nodatavals[self.band - 1]  # type: ignore[no-any-return]

    @property
    def closed(self) -> bool:
        """bool : True if the dataset is closed."""  # noqa: D403
        return self.dataset.closed  # type: ignore[no-any-return]

    def close(self) -> None:
        """
        Close the underlying dataset.

        Has no effect if the dataset is already closed.
        """
        if not self.closed:
            self.dataset.close()

    def __exit__(self, exc_type, exc_value, traceback):  # type: ignore[no-untyped-def]
        self.close()

    def __array__(self) -> np.ndarray:
        return self.dataset.read(self.band)

    def _window_from_slices(
        self, key: slice | tuple[slice, ...]
    ) -> rasterio.windows.Window:
        if isinstance(key, slice):
            row_slice = key
            col_slice = slice(None)
        else:
            row_slice, col_slice = key

        return rasterio.windows.Window.from_slices(
            row_slice, col_slice, height=self.height, width=self.width
        )

    def __getitem__(self, key: slice | tuple[slice, ...], /) -> np.ndarray:
        window = self._window_from_slices(key)
        return self.dataset.read(self.band, window=window)

    def __setitem__(self, key: slice | tuple[slice, ...], value: np.ndarray, /) -> None:
        window = self._window_from_slices(key)
        self.dataset.write(value, self.band, window=window)

    def __repr__(self) -> str:
        clsname = type(self).__name__
        return f"{clsname}(dataset={self.dataset!r}, band={self.band!r})"



from __future__ import annotations

import importlib.resources
import os
import re
import subprocess
from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path

__all__ = [
    "get_snaphu_executable",
    "get_snaphu_version",
    "run_snaphu",
]


@contextmanager
def get_snaphu_executable() -> Generator[Path, None, None]:
    """
    Get the file path of the SNAPHU executable.

    Returns a context manager for use in a ``with`` statement. The SNAPHU executable is
    included in this Python package. If the package was installed as a zip archive, the
    executable will be extracted to a temporary file that is cleaned up upon exiting the
    context manager.

    Yields
    ------
    pathlib.Path
        The file path of the SNAPHU executable.
    """
    files = importlib.resources.files(__package__)
    with importlib.resources.as_file(files / "snaphu") as snaphu:
        yield snaphu


def get_snaphu_version() -> str:
    """
    Get the SNAPHU version, in '<major>.<minor>.<patch>' format.

    Returns
    -------
    str
        The version string reported by the SNAPHU executable.
    """
    # XXX Currently, this runs `snaphu -h` to get the help message and then parses that
    # with a regex to extract the version string. Alternatively, we could:
    #   * parse the `snaphu.h` header file (not zip-safe), or
    #   * simply hard-code the version string here (not easily maintainable)

    # Run `snaphu -h` to get the help message.
    with get_snaphu_executable() as snaphu:
        args = [os.fspath(snaphu), "-h"]
        result = subprocess.run(args, capture_output=True, text=True)  # noqa: PLW1510

        # On success, `snaphu -h` returns 1 (not 0!) and doesn't write to stderr.
        if (result.returncode != 1) or (result.stderr != ""):
            raise subprocess.CalledProcessError(
                result.returncode, result.args, result.stdout, result.stderr
            )

    # Parse the output to extract the version string.
    regex = re.compile(r"^snaphu v(?P<version>[0-9]+(?:\.[0-9]+)*)$", re.MULTILINE)
    match = regex.search(result.stdout)
    if match is None:
        errmsg = "failed to get version string"
        raise RuntimeError(errmsg)

    return match.group("version")


def run_snaphu(config_file: str | os.PathLike[str]) -> None:
    """
    Run SNAPHU with the specified config file.

    Parameters
    ----------
    config_file : path-like
        The file path of a text file storing configuration parameters to pass to SNAPHU.
    """
    if not Path(config_file).is_file():
        errmsg = f"config file not found: {config_file}"
        raise FileNotFoundError(errmsg)

    with get_snaphu_executable() as snaphu:
        args = [os.fspath(snaphu), "-f", os.fspath(config_file)]
        try:
            subprocess.run(args, stderr=subprocess.PIPE, check=True, text=True)
        except subprocess.CalledProcessError as e:
            errmsg = e.stderr.strip()
            raise RuntimeError(errmsg) from e



from typing import Any

from ._dataset import InputDataset, OutputDataset

__all__ = [
    "InputDataset",
    "OutputDataset",
]


def __getattr__(name: str) -> Any:
    if name == "Raster":
        # This module depends on `rasterio` so load it lazily to avoid an ImportError
        # for users that don't need `Raster`.
        from ._raster import Raster

        return Raster

    if name in __all__:
        return globals()[name]

    errmsg = f"module {__name__!r} has no attribute {name!r}"
    raise AttributeError(errmsg)


def __dir__() -> list[str]:
    try:
        from ._raster import Raster
    except ModuleNotFoundError:
        return __all__
    else:
        return sorted([*__all__, "Raster"])



from __future__ import annotations

import itertools
import os
import shutil
from collections.abc import Generator, Iterable, Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from pathlib import Path
from tempfile import mkdtemp

import numpy as np
from numpy.typing import ArrayLike

__all__ = [
    "BlockIterator",
    "ceil_divide",
    "scratch_directory",
]


def ceil_divide(n: ArrayLike, d: ArrayLike) -> np.ndarray:
    """
    Return the smallest integer greater than or equal to the quotient of the inputs.

    Computes integer division of dividend `n` by divisor `d`, rounding up instead of
    truncating.

    Parameters
    ----------
    n : array_like
        The numerator.
    d : array_like
        The denominator.

    Returns
    -------
    q : numpy.ndarray
        The quotient, rounded up to the next integer.
    """
    n = np.asanyarray(n)
    d = np.asanyarray(d)
    return (n + d - np.sign(d)) // d


def as_tuple_of_int(ints: int | Iterable[int]) -> tuple[int, ...]:
    """
    Convert the input to a tuple of ints.

    Parameters
    ----------
    ints : int or iterable of int
        One or more integers.

    Returns
    -------
    out : tuple of int
        Tuple containing the input(s).
    """
    try:
        return (int(ints),)  # type: ignore[arg-type]
    except TypeError:
        return tuple([int(i) for i in ints])  # type: ignore[union-attr]


@dataclass(frozen=True)
class BlockIterator(Iterable[tuple[slice, ...]]):
    """
    An iterable over chunks of an N-dimensional array.

    `BlockIterator` represents a partitioning of a multidimensional array into
    regularly-sized non-overlapping blocks. Each block is represented by an index
    expression (i.e. a tuple of `slice` objects) that can be used to access the
    corresponding block of data from the partitioned array. The full set of blocks spans
    the entire array.

    Iterating over a `BlockIterator` object yields each block in unspecified order.
    """

    shape: tuple[int, ...]
    """tuple of int : The shape of the array to be partitioned into blocks."""
    chunks: tuple[int, ...]
    """
    tuple of int : The shape of a typical block. The last block along each axis may be
    smaller.
    """

    def __init__(self, shape: int | Iterable[int], chunks: int | Iterable[int]):
        """
        Construct a new `BlockIterator` object.

        Parameters
        ----------
        shape : int or iterable of int
            The shape of the array to be partitioned into blocks. Each dimension must be
            > 0.
        chunks : int or iterable of int
            The shape of a typical block. Must be the same length as `shape`. Each chunk
            dimension must be > 0.
        """
        # Normalize `shape` and `chunks` into tuples of ints.
        shape = as_tuple_of_int(shape)
        chunks = as_tuple_of_int(chunks)

        if len(chunks) != len(shape):
            errmsg = (
                "size mismatch: shape and chunks must have the same number of elements,"
                f" instead got len(shape) != len(chunks) ({len(shape)} !="
                f" {len(chunks)})"
            )
            raise ValueError(errmsg)

        if not all(n > 0 for n in shape):
            errmsg = f"shape elements must all be > 0, instead got {shape}"
            raise ValueError(errmsg)
        if any(n <= 0 for n in chunks):
            errmsg = f"chunk elements must all be > 0, instead got {chunks}"
            raise ValueError(errmsg)

        # XXX Workaround for `frozen=True`.
        object.__setattr__(self, "shape", shape)
        object.__setattr__(self, "chunks", chunks)

    def __iter__(self) -> Iterator[tuple[slice, ...]]:
        """
        Iterate over blocks in unspecified order.

        Yields
        ------
        block : tuple of slice
            A tuple of slices that can be used to access the corresponding block of data
            from an array.
        """
        # Number of blocks along each array axis.
        nblocks = ceil_divide(self.shape, self.chunks)

        # Iterate over blocks.
        for block_ind in itertools.product(*[range(n) for n in nblocks]):
            # Get the lower & upper index bounds for the current block.
            start = np.multiply(block_ind, self.chunks)
            stop = np.minimum(start + self.chunks, self.shape)

            # Yield a tuple of slice objects.
            yield tuple(itertools.starmap(slice, zip(start, stop)))


@contextmanager
def scratch_directory(
    dir_: str | os.PathLike[str] | None = None, *, delete: bool = True
) -> Generator[Path, None, None]:
    """
    Context manager that creates a (possibly temporary) file system directory.

    If `dir_` is a path-like object, a directory will be created at the specified
    file system path if it did not already exist. Otherwise, if `dir_` is None, a
    temporary directory will instead be created as though by ``tempfile.mkdtemp()``.

    The directory may be automatically removed from the file system upon exiting the
    context manager.

    Parameters
    ----------
    dir_ : path-like or None, optional
        Scratch directory path. If None, a temporary directory will be created. Defaults
        to None.
    delete : bool, optional
        If True, the directory and its contents are recursively removed from the
        file system upon exiting the context manager. Defaults to True.

    Yields
    ------
    pathlib.Path
        Scratch directory path. If `delete` was True, the directory will be removed from
        the file system upon exiting the context manager scope.
    """
    if dir_ is None:
        scratchdir = Path(mkdtemp())
    else:
        scratchdir = Path(dir_)
        scratchdir.mkdir(parents=True, exist_ok=True)

    yield scratchdir

    if delete:
        shutil.rmtree(scratchdir)



import itertools
import shutil
import tempfile
from collections.abc import Iterable
from pathlib import Path

import numpy as np
import pytest

from snaphu._util import BlockIterator, ceil_divide, scratch_directory


class TestCeilDivide:
    def test_positive(self):
        assert ceil_divide(3, 2) == 2
        assert ceil_divide(4, 2) == 2
        assert ceil_divide(1, 1_000_000) == 1

    def test_negative(self):
        assert ceil_divide(-3, 2) == -1
        assert ceil_divide(3, -2) == -1
        assert ceil_divide(-3, -2) == 2

        assert ceil_divide(-4, 2) == -2
        assert ceil_divide(4, -2) == -2
        assert ceil_divide(-4, -2) == 2

    def test_zero(self):
        assert ceil_divide(0, 1) == 0
        assert ceil_divide(-0, 1) == 0

    def test_divide_by_zero(self):
        with pytest.warns(RuntimeWarning) as record:
            ceil_divide(1, 0)

        assert len(record) == 1
        warning = record[0].message.args[0]
        assert "divide by zero encountered" in warning

    def test_arraylike(self):
        result = ceil_divide([1, 2, 3, 4, 5], 2)
        expected = [1, 1, 2, 2, 3]
        np.testing.assert_array_equal(result, expected)


class TestBlockIterator:
    @pytest.fixture
    def blocks2d(self) -> BlockIterator:
        return BlockIterator(shape=(100, 101), chunks=(25, 34))

    def test_is_iterable(self, blocks2d: BlockIterator):
        assert isinstance(blocks2d, Iterable)

    def test_attrs(self, blocks2d: BlockIterator):
        assert blocks2d.shape == (100, 101)
        assert blocks2d.chunks == (25, 34)

    def test_nblocks(self, blocks2d: BlockIterator):
        nblocks = len(list(blocks2d))
        assert nblocks == 12

    def test_iter(self, blocks2d: BlockIterator):
        arr = np.zeros(blocks2d.shape, dtype=np.int32)
        for block in blocks2d:
            arr[block] += 1
        np.testing.assert_array_equal(arr, 1)

    def test_blocks1d(self):
        blocks = BlockIterator(shape=99, chunks=25)
        assert blocks.shape == (99,)
        assert blocks.chunks == (25,)

        starts = [0, 25, 50, 75]
        stops = [25, 50, 75, 99]
        for block, start, stop in itertools.zip_longest(blocks, starts, stops):
            (slice_,) = block
            assert slice_.start == start
            assert slice_.stop == stop
            assert slice_.step is None

    def test_shape_chunks_mismatch(self):
        pattern = (
            "^size mismatch: shape and chunks must have the same number of elements,"
            r" instead got len\(shape\) != len\(chunks\) \(2 != 3\)$"
        )
        with pytest.raises(ValueError, match=pattern):
            BlockIterator(shape=(300, 400), chunks=(3, 4, 5))

    def test_bad_shape(self):
        pattern = r"^shape elements must all be > 0, instead got \(100, -1\)$"
        with pytest.raises(ValueError, match=pattern):
            BlockIterator(shape=(100, -1), chunks=(10, 10))

    def test_bad_chunks(self):
        pattern = r"^chunk elements must all be > 0, instead got \(10, -1\)$"
        with pytest.raises(ValueError, match=pattern):
            BlockIterator(shape=(100, 100), chunks=(10, -1))

    def test_repr(self, blocks2d: BlockIterator):
        assert repr(blocks2d) == "BlockIterator(shape=(100, 101), chunks=(25, 34))"


class TestScratchDirectory:
    def test_temp_directory(self):
        with scratch_directory() as scratchdir:
            assert scratchdir.is_dir()
        assert not scratchdir.exists()

    def test_create_directory(self):
        # Create a temporary parent directory.
        with tempfile.TemporaryDirectory() as tmpdir:
            # The file path to create the scratch directory at.
            path = Path(tmpdir) / "a/s/d/f"
            assert not path.is_dir()

            # Create a new scratch directory.
            with scratch_directory(path) as scratchdir:
                # Check that a directory was created in the right place.
                assert scratchdir == path
                assert scratchdir.is_dir()

    def test_existing_directory(self):
        with tempfile.TemporaryDirectory() as tmpdir:  # noqa: SIM117
            with scratch_directory(tmpdir, delete=False) as scratchdir:
                assert scratchdir == Path(tmpdir)

    @pytest.mark.parametrize("delete", [True, False])
    def test_delete(self, *, delete: bool):
        with tempfile.TemporaryDirectory() as tmpdir:
            # Create a scratch directory at the specified path.
            path = Path(tmpdir) / "asdf"
            with scratch_directory(path, delete=delete) as scratchdir:
                assert scratchdir.is_dir()

            # Check that the scratch directory was removed if `delete` was True.
            if delete:
                assert not scratchdir.exists()
            else:
                assert scratchdir.is_dir()

    @pytest.mark.parametrize("delete", [True, False])
    def test_tempdir_delete(self, *, delete: bool):
        try:
            # Create a scratch directory using `mkdtemp()`.
            with scratch_directory(delete=delete) as scratchdir:
                assert scratchdir.is_dir()

            # Check that the scratch directory was removed if `delete` was True.
            if delete:
                assert not scratchdir.exists()
            else:
                assert scratchdir.exists()
        finally:
            # Clean up the directory if it wasn't automatically removed.
            if scratchdir.is_dir():
                shutil.rmtree(scratchdir)



from __future__ import annotations

from typing import Protocol, runtime_checkable

import numpy as np
from numpy.typing import ArrayLike

__all__ = [
    "InputDataset",
    "OutputDataset",
]


@runtime_checkable
class InputDataset(Protocol):
    """
    An array-like interface for reading input datasets.

    `InputDataset` defines the abstract interface that types must conform to in order
    to be valid inputs to the ``snaphu.unwrap()`` function. Such objects must export
    NumPy-like `dtype`, `shape`, and `ndim` attributes and must support NumPy-style
    slice-based indexing.

    See Also
    --------
    OutputDataset
    """

    @property
    def dtype(self) -> np.dtype:
        """numpy.dtype : Data-type of the array's elements."""

    @property
    def shape(self) -> tuple[int, ...]:
        """tuple of int : Tuple of array dimensions."""  # noqa: D403

    @property
    def ndim(self) -> int:
        """int : Number of array dimensions."""  # noqa: D403

    def __getitem__(self, key: slice | tuple[slice, ...], /) -> ArrayLike:
        """Read a block of data."""


@runtime_checkable
class OutputDataset(Protocol):
    """
    An array-like interface for writing output datasets.

    `OutputDataset` defines the abstract interface that types must conform to in order
    to be valid outputs of the ``snaphu.unwrap()`` function. Such objects must export
    NumPy-like `dtype`, `shape`, and `ndim` attributes and must support NumPy-style
    slice-based indexing.

    See Also
    --------
    InputDataset
    """

    @property
    def dtype(self) -> np.dtype:
        """numpy.dtype : Data-type of the array's elements."""

    @property
    def shape(self) -> tuple[int, ...]:
        """tuple of int : Tuple of array dimensions."""  # noqa: D403

    @property
    def ndim(self) -> int:
        """int : Number of array dimensions."""  # noqa: D403

    def __setitem__(self, key: slice | tuple[slice, ...], value: np.ndarray, /) -> None:
        """Write a block of data."""



from __future__ import annotations

import os
import textwrap
from pathlib import Path
from tempfile import mkstemp
from typing import cast, overload

import numpy as np

from ._snaphu import run_snaphu
from ._util import BlockIterator, scratch_directory
from .io import InputDataset, OutputDataset

__all__ = [
    "unwrap",
]


def check_shapes(
    igram: InputDataset,
    corr: InputDataset,
    unw: OutputDataset,
    conncomp: OutputDataset,
    mask: InputDataset | None = None,
) -> None:
    """
    Check that the arguments are 2-D arrays with matching shapes.

    Parameters
    ----------
    igram : snaphu.io.InputDataset
        The input interferogram. Must be a 2-D array.
    corr : snaphu.io.InputDataset
        The input coherence. Must be a 2-D array with the same shape as `igram`.
    unw : snaphu.io.OutputDataset
        The output unwrapped phase. Must be a 2-D array with the same shape as `igram`.
    conncomp : snaphu.io.OutputDataset
        The output connected component labels. Must be a 2-D array with the same shape
        as `igram`.
    mask : snaphu.io.InputDataset or None, optional
        An optional binary mask of valid samples. If not None, it must be a 2-D array
        with the same shape as `igram`. Defaults to None.

    Raises
    ------
    ValueError
        If any array is not 2-D or has a different shape than the others.
    """
    if igram.ndim != 2:
        errmsg = f"igram must be 2-D, instead got ndim={igram.ndim}"
        raise ValueError(errmsg)

    def raise_shape_mismatch_error(
        name: str,
        arr: InputDataset | OutputDataset,
    ) -> None:
        errmsg = (
            f"shape mismatch: {name} and igram must have the same shape, instead got"
            f" {name}.shape={arr.shape} and {igram.shape=}"
        )
        raise ValueError(errmsg)

    if corr.shape != igram.shape:
        raise_shape_mismatch_error("corr", corr)
    if unw.shape != igram.shape:
        raise_shape_mismatch_error("unw", unw)
    if conncomp.shape != igram.shape:
        raise_shape_mismatch_error("conncomp", conncomp)
    if (mask is not None) and (mask.shape != igram.shape):
        raise_shape_mismatch_error("mask", mask)


def check_dtypes(
    igram: InputDataset,
    corr: InputDataset,
    unw: OutputDataset,
    conncomp: OutputDataset,
    mask: InputDataset | None = None,
) -> None:
    """
    Check that the arguments have valid datatypes.

    Parameters
    ----------
    igram : snaphu.io.InputDataset
        The input interferogram. Must be a complex-valued array.
    corr : snaphu.io.InputDataset
        The input coherence. Must be a real-valued array.
    unw : snaphu.io.OutputDataset
        The output unwrapped phase. Must be a real-valued array.
    conncomp : snaphu.io.OutputDataset
        The output connected component labels. Must be an integer-valued array.
    mask : snaphu.io.InputDataset or None, optional
        An optional binary mask of valid samples. If not None, it must be a boolean or
        8-bit integer array.

    Raises
    ------
    TypeError
        If any array has an unexpected datatype.
    """
    if not np.issubdtype(igram.dtype, np.complexfloating):
        errmsg = (
            f"igram must be a complex-valued array, instead got dtype={igram.dtype}"
        )
        raise TypeError(errmsg)

    if not np.issubdtype(corr.dtype, np.floating):
        errmsg = f"corr must be a real-valued array, instead got dtype={corr.dtype}"
        raise TypeError(errmsg)

    if not np.issubdtype(unw.dtype, np.floating):
        errmsg = f"unw must be a real-valued array, instead got dtype={unw.dtype}"
        raise TypeError(errmsg)

    if not np.issubdtype(conncomp.dtype, np.integer):
        errmsg = (
            "conncomp must be an integer-valued array, instead got"
            f" dtype={conncomp.dtype}"
        )
        raise TypeError(errmsg)

    if (
        (mask is not None)
        and (mask.dtype != np.bool_)
        and (mask.dtype != np.uint8)
        and (mask.dtype != np.int8)
    ):
        errmsg = (
            "mask must be a boolean or 8-bit integer array (or None), instead got"
            f" dtype={mask.dtype}"
        )
        raise TypeError(errmsg)


def normalize_and_validate_tiling_params(
    ntiles: tuple[int, int],
    tile_overlap: int | tuple[int, int],
    nproc: int,
) -> tuple[tuple[int, int], tuple[int, int], int]:
    """
    Normalize and validate inputs related to tiling and multiprocessing.

    Parameters
    ----------
    ntiles : (int, int)
        Number of tiles along the row/column directions.
    tile_overlap : int or (int, int)
        Overlap, in pixels, between neighboring tiles.
    nproc : int
        Maximum number of child processes to spawn for parallel tile unwrapping.

    Returns
    -------
    ntiles : (int, int)
        `ntiles` normalized to a pair of positive integers.
    tile_overlap : (int, int)
        `tile_overlap` normalized to a pair of nonnegative integers.
    nproc : int
        `nproc` as a positive integer.
    """
    # Normalize `ntiles` to a tuple and ensure its contents are two positive-valued
    # integers.
    ntiles = tuple(ntiles)  # type: ignore[assignment]
    if len(ntiles) != 2:
        errmsg = f"ntiles must be a pair of ints, instead got {ntiles=}"
        raise ValueError(errmsg)
    if not all(n >= 1 for n in ntiles):
        errmsg = f"ntiles may not contain negative or zero values, got {ntiles=}"
        raise ValueError(errmsg)

    # If `tile_overlap` is iterable, ensure it's a tuple. Otherwise, assume it's a
    # single integer.
    try:
        tile_overlap = tuple(tile_overlap)  # type: ignore[arg-type,assignment]
    except TypeError:
        tile_overlap = (tile_overlap, tile_overlap)  # type: ignore[assignment]

    # Convince static type checkers that `tile_overlap` is now a pair of ints.
    tile_overlap = cast(tuple[int, int], tile_overlap)

    # Ensure the contents of `tile_overlap` are two nonnegative integers.
    if len(tile_overlap) != 2:
        errmsg = (
            f"tile_overlap must be an int or pair of ints, instead got {tile_overlap=}"
        )
        raise ValueError(errmsg)
    if not all(n >= 0 for n in tile_overlap):
        errmsg = f"tile_overlap may not contain negative values, got {tile_overlap=}"
        raise ValueError(errmsg)

    # If `nproc` is less than 1, use all available processors. Fall back to serial
    # execution if the number of available processors cannot be determined.
    if nproc < 1:
        nproc = os.cpu_count() or 1

    return ntiles, tile_overlap, nproc


def copy_blockwise(
    src: InputDataset,
    dst: OutputDataset,
    chunks: tuple[int, int] = (1024, 1024),
    *,
    nan_to_zero: bool = False,
) -> None:
    """
    Copy the contents of `src` to `dst` block-by-block.

    Parameters
    ----------
    src : snaphu.io.InputDataset
        Source dataset.
    dst : snaphu.io.OutputDataset
        Destination dataset.
    chunks : (int, int), optional
        Block dimensions. Defaults to (1024, 1024).
    nan_to_zero : bool, optional
        If True, replace Not a Number (NaN) values in the array with zeros in the
        output. Defaults to False.
    """
    shape = src.shape
    if dst.shape != shape:
        errmsg = (
            "shape mismatch: src and dst must have the same shape, instead got"
            f" {src.shape=} and {dst.shape=}"
        )
        raise ValueError(errmsg)

    for block in BlockIterator(shape, chunks):
        if nan_to_zero:
            nan_mask = np.isnan(src[block])
            dst[block] = np.where(nan_mask, 0.0, src[block])
        else:
            dst[block] = src[block]


def regrow_conncomp_from_unw(
    unw_file: str | os.PathLike[str],
    corr_file: str | os.PathLike[str],
    conncomp_file: str | os.PathLike[str],
    line_length: int,
    nlooks: float,
    cost: str,
    mask_file: str | os.PathLike[str] | None = None,
    scratchdir: str | os.PathLike[str] | None = None,
) -> None:
    """
    Run SNAPHU to regrow connected components from an unwrapped input.

    This is particularly useful if SNAPHU was initially run in tiled mode, resulting in
    a set of connected components that are disjoint across tile boundaries. The
    connected components will be recomputed as though by a single tile, so that
    components are no longer delimited by the extents of each tile.

    This function does not compute a new unwrapped solution.

    Parameters
    ----------
    unw_file : path-like
        The input unwrapped phase file path.
    corr_file : path-like
        The input coherence file path.
    conncomp_file : path-like
        The output connected component labels file path.
    line_length : int
        The line length, in samples, of the input & output arrays.
    nlooks : float
        The equivalent number of independent looks used to form the sample coherence.
    cost : str
        Statistical cost mode.
    mask_file : path-like or None, optional
        An optional file path of a byte mask file. If None, no mask is applied. Defaults
        to None.
    scratchdir : path-like or None, optional
        The scratch directory where the config file will be written. If None, a
        default temporary directory is chosen as though by ``tempfile.gettempdir()``.
        Defaults to None.
    """
    # In REGROWCONNCOMPS mode, SNAPHU recomputes the cost arrays (but does not
    # re-unwrap), so we should pass in all parameters necessary to compute costs, but
    # don't need to pass an INITMETHOD, for example.
    config = textwrap.dedent(f"""\
        REGROWCONNCOMPS TRUE
        INFILE {os.fspath(unw_file)}
        INFILEFORMAT FLOAT_DATA
        UNWRAPPEDINFILEFORMAT FLOAT_DATA
        CORRFILE {os.fspath(corr_file)}
        CORRFILEFORMAT FLOAT_DATA
        CONNCOMPFILE {os.fspath(conncomp_file)}
        CONNCOMPOUTTYPE UINT
        LINELENGTH {line_length}
        NCORRLOOKS {nlooks}
        STATCOSTMODE {cost.upper()}
    """)
    if mask_file is not None:
        config += f"BYTEMASKFILE {os.fspath(mask_file)}\n"

    # Write config parameters to file. The config file should have a descriptive name to
    # disambiguate it from the config file used for unwrapping.
    _, config_file = mkstemp(
        dir=scratchdir, prefix="snaphu-regrow-conncomps.config.", suffix=".txt"
    )
    Path(config_file).write_text(config)

    # Run SNAPHU in REGROWCONNCOMPS mode to generate new connected component labels.
    run_snaphu(config_file)


@overload
def unwrap(
    igram: InputDataset,
    corr: InputDataset,
    nlooks: float,
    cost: str = "smooth",
    init: str = "mcf",
    *,
    mask: InputDataset | None = None,
    ntiles: tuple[int, int] = (1, 1),
    tile_overlap: int | tuple[int, int] = 0,
    nproc: int = 1,
    regrow_conncomps: bool = True,
    scratchdir: str | os.PathLike[str] | None = None,
    delete_scratch: bool = True,
    unw: OutputDataset,
    conncomp: OutputDataset,
) -> tuple[OutputDataset, OutputDataset]: ...  # pragma: no cover


# If `unw` and `conncomp` aren't specified, return the outputs as two NumPy arrays.
@overload
def unwrap(
    igram: InputDataset,
    corr: InputDataset,
    nlooks: float,
    cost: str = "smooth",
    init: str = "mcf",
    *,
    mask: InputDataset | None = None,
    ntiles: tuple[int, int] = (1, 1),
    tile_overlap: int | tuple[int, int] = 0,
    nproc: int = 1,
    regrow_conncomps: bool = True,
    scratchdir: str | os.PathLike[str] | None = None,
    delete_scratch: bool = True,
) -> tuple[np.ndarray, np.ndarray]: ...  # pragma: no cover


def unwrap(  # type: ignore[no-untyped-def]
    igram,
    corr,
    nlooks,
    cost="smooth",
    init="mcf",
    *,
    mask=None,
    ntiles=(1, 1),
    tile_overlap=0,
    nproc=1,
    regrow_conncomps=True,
    scratchdir=None,
    delete_scratch=True,
    unw=None,
    conncomp=None,
):
    r"""
    Unwrap an interferogram using SNAPHU.

    Performs 2-D phase unwrapping using the Statistical-Cost, Network-Flow Algorithm for
    Phase Unwrapping (SNAPHU)\ [1]_. The algorithm produces a Maximum a Posteriori (MAP)
    estimate of the unwrapped phase field by (approximately) solving a nonlinear network
    flow optimization problem.

    The outputs include the unwrapped phase and a corresponding array of connected
    component labels. Each connected component is a region of pixels in the solution
    that is believed to have been unwrapped in an internally self-consistent manner\
    [2]_. Each distinct region is assigned a unique positive integer label. Pixels not
    belonging to any component are assigned a label of zero.

    Parameters
    ----------
    igram : snaphu.io.InputDataset
        The input interferogram. A 2-D complex-valued array. Not a Number (NaN) values
        in the array will be replaced with zeros.
    corr : snaphu.io.InputDataset
        The sample coherence magnitude. Must be a floating-point array with the same
        dimensions as the input interferogram. Valid coherence values are in the range
        [0, 1]. NaN values in the array will be replaced with zeros.
    nlooks : float
        The equivalent number of independent looks used to form the sample coherence. An
        estimate of the number of statistically independent samples averaged in the
        multilooked data, taking into account spatial correlation due to
        oversampling/filtering (see `Notes`_).
    cost : {'defo', 'smooth'}, optional
        Statistical cost mode. Defaults to 'smooth'.
    init : {'mst', 'mcf'}, optional
        Algorithm used for initialization of unwrapped phase gradients.
        Supported algorithms include Minimum Spanning Tree ('mst') and Minimum
        Cost Flow ('mcf'). Defaults to 'mcf'.
    mask : snaphu.io.InputDataset or None, optional
        Binary mask of valid pixels. Zeros in this raster indicate interferogram
        pixels that should be masked out. If provided, it must have the same dimensions
        as the input interferogram and boolean or 8-bit integer datatype. Defaults to
        None.
    ntiles : (int, int), optional
        Number of tiles along the row/column directions. If `ntiles` is (1, 1), then the
        interferogram will be unwrapped as a single tile. Increasing the number of tiles
        may improve runtime and reduce peak memory utilization, but may also introduce
        tile boundary artifacts in the unwrapped result. Defaults to (1, 1).
    tile_overlap : int or (int, int), optional
        Overlap, in pixels, between neighboring tiles. Increasing overlap may help to
        avoid phase discontinuities between tiles. If `tile_overlap` is a scalar
        integer, the number of overlapping rows and columns will be the same. Defaults
        to 0.
    nproc : int, optional
        Maximum number of child processes to spawn for parallel tile unwrapping. If
        `nproc` is less than 1, use all available processors. Defaults to 1.
    regrow_conncomps : bool, optional
        If True, the connected component labels will be re-computed using a single tile
        after first unwrapping with multiple tiles. This option is disregarded when
        `ntiles` is (1, 1). Defaults to True.
    scratchdir : path-like or None, optional
        Scratch directory where intermediate processing artifacts are written.
        If the specified directory does not exist, it will be created. If None,
        a temporary directory will be created as though by
        ``tempfile.TemporaryDirectory()``. Defaults to None.
    delete_scratch : bool, optional
        If True, the scratch directory will be automatically removed from the file
        system when the function exits. Otherwise, the scratch directory will be
        preserved. Defaults to True.
    unw : snaphu.io.OutputDataset or None, optional
        An optional output dataset to store the unwrapped phase, in radians. If
        provided, it must have the same dimensions as the input interferogram and
        floating-point datatype. If None, an output array will be allocated internally.
        Defaults to None.
    conncomp : snaphu.io.OutputDataset or None, optional
        An optional output dataset to store the connected component labels. If provided,
        it must have the same dimensions as the input interferogram and any integer
        datatype. If None, an output array will be allocated internally. Defaults to
        None.

    Returns
    -------
    unw : snaphu.io.OutputDataset
        The output unwrapped phase, in radians.
    conncomp : snaphu.io.OutputDataset
        The output connected component labels.

    Notes
    -----
    An estimate of the equivalent number of independent looks may be obtained by

    .. math:: n_e = k_r k_a \frac{d_r d_a}{\rho_r \rho_a}

    where :math:`k_r` and :math:`k_a` are the number of looks in range and azimuth,
    :math:`d_r` and :math:`d_a` are the (single-look) sample spacing in range and
    azimuth, and :math:`\rho_r` and :math:`\rho_a are the range and azimuth resolution.

    References
    ----------
    .. [1] C. W. Chen and H. A. Zebker, "Two-dimensional phase unwrapping with use of
       statistical models for cost functions in nonlinear optimization," Journal of the
       Optical Society of America A, vol. 18, pp. 338-351 (2001).
    .. [2] C. W. Chen and H. A. Zebker, "Phase unwrapping for large SAR interferograms:
       Statistical segmentation and generalized network models," IEEE Transactions on
       Geoscience and Remote Sensing, vol. 40, pp. 1709-1719 (2002).
    """
    # If the `unw` and/or `conncomp` output datasets were not provided, allocate arrays
    # to store these outputs.
    if unw is None:
        unw = np.zeros(shape=igram.shape, dtype=np.float32)
    if conncomp is None:
        conncomp = np.zeros(shape=igram.shape, dtype=np.uint32)

    # Ensure that input & output datasets have valid dimensions and datatypes.
    check_shapes(igram, corr, unw, conncomp, mask)
    check_dtypes(igram, corr, unw, conncomp, mask)

    if nlooks < 1.0:
        errmsg = f"nlooks must be >= 1, instead got {nlooks}"
        raise ValueError(errmsg)

    if cost == "topo":
        errmsg = "'topo' cost mode is not currently supported"
        raise NotImplementedError(errmsg)

    cost_modes = {"defo", "smooth"}
    if cost not in cost_modes:
        errmsg = f"cost mode must be in {cost_modes}, instead got {cost!r}"
        raise ValueError(errmsg)

    init_methods = {"mst", "mcf"}
    if init not in init_methods:
        errmsg = f"init method must be in {init_methods}, instead got {init!r}"
        raise ValueError(errmsg)

    # Validate inputs related to tiling and coerce them to the expected types.
    ntiles, tile_overlap, nproc = normalize_and_validate_tiling_params(
        ntiles=ntiles, tile_overlap=tile_overlap, nproc=nproc
    )

    with scratch_directory(scratchdir, delete=delete_scratch) as dir_:
        # Create a raw binary file in the scratch directory for the interferogram and
        # copy the input data to it. (`mkstemp` is used to avoid data races in case the
        # same scratch directory was used for multiple SNAPHU processes.)
        _, tmp_igram = mkstemp(dir=dir_, prefix="snaphu.igram.", suffix=".c8")
        tmp_igram_mmap = np.memmap(tmp_igram, dtype=np.complex64, shape=igram.shape)
        copy_blockwise(igram, tmp_igram_mmap, nan_to_zero=True)

        # Copy the input coherence data to a raw binary file in the scratch directory.
        _, tmp_corr = mkstemp(dir=dir_, prefix="snaphu.corr.", suffix=".f4")
        tmp_corr_mmap = np.memmap(tmp_corr, dtype=np.float32, shape=corr.shape)
        copy_blockwise(corr, tmp_corr_mmap, nan_to_zero=True)

        # If a mask was provided, copy the mask data to a raw binary file in the scratch
        # directory.
        if mask is None:
            tmp_mask = None
        else:
            _, tmp_mask = mkstemp(dir=dir_, prefix="snaphu.mask.", suffix=".u1")
            tmp_mask_mmap = np.memmap(tmp_mask, dtype=np.bool_, shape=mask.shape)
            copy_blockwise(mask, tmp_mask_mmap)

        # Create files in the scratch directory for SNAPHU outputs.
        _, tmp_unw = mkstemp(dir=dir_, prefix="snaphu.unw.", suffix=".f4")
        _, tmp_conncomp = mkstemp(dir=dir_, prefix="snaphu.conncomp.", suffix=".u4")

        config = textwrap.dedent(f"""\
            INFILE {tmp_igram}
            INFILEFORMAT COMPLEX_DATA
            CORRFILE {tmp_corr}
            CORRFILEFORMAT FLOAT_DATA
            OUTFILE {tmp_unw}
            OUTFILEFORMAT FLOAT_DATA
            CONNCOMPFILE {tmp_conncomp}
            CONNCOMPOUTTYPE UINT
            LINELENGTH {igram.shape[1]}
            NCORRLOOKS {nlooks}
            STATCOSTMODE {cost.upper()}
            INITMETHOD {init.upper()}
            NTILEROW {ntiles[0]}
            NTILECOL {ntiles[1]}
            ROWOVRLP {tile_overlap[0]}
            COLOVRLP {tile_overlap[1]}
            NPROC {nproc}
        """)
        if mask is not None:
            config += f"BYTEMASKFILE {tmp_mask}\n"

        # Write config parameters to file.
        _, config_file = mkstemp(dir=dir_, prefix="snaphu.config.", suffix=".txt")
        Path(config_file).write_text(config)

        # Run SNAPHU with the specified parameters.
        run_snaphu(config_file)

        # Optionally re-run SNAPHU to regrow connected components from the unwrapped
        # phase as though in single-tile mode, overwriting the original connected
        # components file. This step should have no effect if SNAPHU was previously run
        # in single-tile mode, so skip it in that case.
        single_tile = ntiles == (1, 1)
        if (not single_tile) and regrow_conncomps:
            regrow_conncomp_from_unw(
                unw_file=tmp_unw,
                corr_file=tmp_corr,
                conncomp_file=tmp_conncomp,
                line_length=igram.shape[1],
                nlooks=nlooks,
                cost=cost,
                scratchdir=dir_,
                mask_file=tmp_mask,
            )

        # Get the output unwrapped phase data.
        tmp_unw_mmap = np.memmap(tmp_unw, dtype=np.float32, shape=unw.shape)
        copy_blockwise(tmp_unw_mmap, unw)

        # Get the output connected component labels.
        tmp_cc_mmap = np.memmap(tmp_conncomp, dtype=np.uint32, shape=conncomp.shape)
        copy_blockwise(tmp_cc_mmap, conncomp)

    return unw, conncomp



import os
import re
import tempfile
from pathlib import Path

import pytest

import snaphu
from snaphu._snaphu import get_snaphu_executable, run_snaphu


def test_get_snaphu_executable():
    # Check that the file exists and is executable.
    with get_snaphu_executable() as snaphu_exe:
        assert Path(snaphu_exe).is_file()
        assert os.access(snaphu_exe, os.X_OK)


def test_get_snaphu_version():
    version = snaphu.get_snaphu_version()

    # The result should be a string in '<major>.<minor>.<patch>' format.
    regex = re.compile(r"^[0-9]+(\.[0-9]+){2}$")
    assert regex.fullmatch(version)


class TestRunSnaphu:
    def test_empty_config(self):
        # Test running SNAPHU with an empty config file.
        with tempfile.NamedTemporaryFile() as config_file:
            pattern = r"^not enough input arguments\.\s+type snaphu -h for help$"
            with pytest.raises(RuntimeError, match=pattern):
                run_snaphu(config_file.name)

    def test_missing_config(self):
        config_file = "/this/file/does/not/exist"
        assert not Path(config_file).is_file()

        # Should raise a `FileNotFoundError` if the config file did not exist.
        pattern = r"^config file not found"
        with pytest.raises(FileNotFoundError, match=pattern):
            run_snaphu(config_file)
